# The Principal code is taken from https://github.com/rocklabs-io/ic-py/blob/main/ic/principal.py

# MIT License

# Copyright (c) 2021 Rocklabs

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# principal type: https://github.com/dfinity/ic-types/blob/main/src/principal.rs

import zlib
import math
import base64
import hashlib
from enum import Enum

from typing import Union

CRC_LENGTH_IN_BYTES = 4
HASH_LENGTH_IN_BYTES = 28
MAX_LENGTH_IN_BYTES = 29


class PrincipalClass(Enum):
    OpaqueId = 1
    SelfAuthenticating = 2
    DerivedId = 3
    Anonymous = 4
    # Unassigned


class Principal:
    def __init__(self, bytes: bytes = b""):
        self._len = len(bytes)
        self._bytes = bytes
        self.hex = str(self._bytes.hex()).upper()
        self._isPrincipal = True

    @staticmethod
    def management_canister():
        return Principal()

    @staticmethod
    def self_authenticating(pubkey: Union[str, bytes]):
        if isinstance(pubkey, str):
            pubkey = bytes.fromhex(pubkey)
        hash_ = hashlib.sha224(pubkey).digest()
        hash_ += bytes([PrincipalClass.SelfAuthenticating.value])
        return Principal(bytes=hash_)

    @staticmethod
    def anonymous():
        return Principal(bytes=b"\x04")

    @property
    def len(self):
        return self._len

    @property
    def bytes(self):
        return self._bytes

    @property
    def isPrincipal(self):
        return self._isPrincipal

    @staticmethod
    def from_str(s: str):
        s1 = s.replace("-", "")
        pad_len = math.ceil(len(s1) / 8) * 8 - len(s1)
        b = base64.b32decode(s1.upper().encode() + b"=" * pad_len)
        if len(b) < CRC_LENGTH_IN_BYTES:
            raise Exception("principal length error")
        p = Principal(bytes=b[CRC_LENGTH_IN_BYTES:])
        if not p.to_str() == s:
            raise Exception("principal format error")
        return p

    @staticmethod
    def from_hex(s: str):
        return Principal(bytes.fromhex(s.lower()))

    def to_str(self):
        checksum = zlib.crc32(self._bytes) & 0xFFFFFFFF
        b = b""
        b += checksum.to_bytes(CRC_LENGTH_IN_BYTES, byteorder="big")
        b += self.bytes
        s = base64.b32encode(b).decode("utf-8").lower().replace("=", "")
        ret = ""
        while len(s) > 5:
            ret += s[:5]
            ret += "-"
            s = s[5:]
        ret += s
        return ret

    def to_account_id(self, sub_account: int = 0):
        return AccountIdentifier.new(self, sub_account)

    def __repr__(self):
        return "Principal(" + self.to_str() + ")"

    def __str__(self):
        return self.to_str()


class AccountIdentifier:
    def __init__(self, hash: bytes) -> None:
        assert len(hash) == 32
        self._hash = hash

    def to_str(self):
        return "0x" + self._hash.hex()

    def __repr__(self):
        return "Account(" + self.to_str() + ")"

    def __str__(self):
        return self.to_str()

    @property
    def bytes(self) -> bytes:
        return self._hash

    @staticmethod
    def new(principal: Principal, sub_account: int = 0):
        sha224 = hashlib.sha224()
        sha224.update(b"\x0Aaccount-id")
        sha224.update(principal.bytes)

        sub_account = sub_account.to_bytes(32, byteorder="big")  # type: ignore
        sha224.update(sub_account)  # type: ignore
        hash = sha224.digest()
        checksum = zlib.crc32(hash) & 0xFFFFFFFF
        account = checksum.to_bytes(CRC_LENGTH_IN_BYTES, byteorder="big") + hash
        return AccountIdentifier(account)



import sys
from typing import (
    Annotated,
    Any,
    Callable,
    Generator,
    Generic,
    NoReturn,
    Optional,
    ParamSpec,
    TypedDict,
    TypeVar,
    TypeAlias,
    Union,
)

# TODO I think we can simplify this just like we're doing with canisters
from .compiler.custom_modules.principal import Principal as PrincipalRenamed

__version__ = "0.8.19"
__rust_version__ = "1.88.0"

Principal = PrincipalRenamed

int64 = int
int32 = int
int16 = int
int8 = int

nat = int
nat64 = int
nat32 = int
nat16 = int
nat8 = int

float64 = float
float32 = float

text = str

T = TypeVar("T")
Opt = Optional[T]
Manual = Optional[T]
Alias = Annotated[T, None]

Tuple = tuple
Vec = list

Record = TypedDict
Variant = TypedDict

blob = bytes

null: TypeAlias = None
void: TypeAlias = None

reserved = Any
empty: TypeAlias = NoReturn  # TODO in Python 3.11 I believe there is a Never type

Async = Generator[Any, Any, T]

TimerId = Alias[nat64]
Duration = Alias[nat64]


class GuardResult(Variant, total=False):
    Ok: null
    Err: str


GuardType = Callable[..., GuardResult]


def query(
    _func: Optional[Callable[..., Any]] = None, *, guard: Optional[GuardType] = None
) -> Callable[..., Any]:
    def decorator(func: Callable[..., Any]):
        return func

    if _func is None:
        return decorator
    else:
        return decorator(_func)


def update(
    _func: Optional[Callable[..., Any]] = None, *, guard: Optional[GuardType] = None
) -> Callable[..., Any]:
    def decorator(func: Callable[..., Any]):
        return func

    if _func is None:
        return decorator
    else:
        return decorator(_func)


def canister(cls: T) -> T:
    return cls


def init(func: object):
    return func


def heartbeat(
    _func: Optional[Callable[..., Any]] = None, *, guard: Optional[GuardType] = None
) -> Callable[..., Any]:
    def decorator(func: Callable[..., Any]):
        return func

    if _func is None:
        return decorator
    else:
        return decorator(_func)


def pre_upgrade(
    _func: Optional[Callable[..., Any]] = None, *, guard: Optional[GuardType] = None
) -> Callable[..., Any]:
    def decorator(func: Callable[..., Any]):
        return func

    if _func is None:
        return decorator
    else:
        return decorator(_func)


def post_upgrade(func: object):
    return func


def inspect_message(
    _func: Optional[Callable[..., Any]] = None, *, guard: Optional[GuardType] = None
) -> Callable[..., Any]:
    def decorator(func: Callable[..., Any]):
        return func

    if _func is None:
        return decorator
    else:
        return decorator(_func)


# TODO need service
Query = Callable
Update = Callable
Oneway = Callable


class CallResult(Generic[T]):
    Ok: T
    Err: Optional[str]

    def __init__(self, Ok: T, Err: str):
        self.Ok = Ok
        self.Err = Err

    def notify(self) -> "NotifyResult": ...

    def with_cycles(self, cycles: nat64) -> "CallResult[T]": ...

    def with_cycles128(self, cycles: nat) -> "CallResult[T]": ...


# TODO Once RustPython supports Python 3.11, we can use the below and unify CallResult with the other Variants
# TODO The problem is that you can't really use generics with TypedDict yet: https://github.com/python/cpython/issues/89026
# TODO We could also consider a hack where we remove all references to CallResult before runtime, since this is really an analysis-time consideration
# class CallResult(Variant, Generic[T], total=False):
#     Ok: T
#     Err: str


class RejectionCode(Variant, total=False):
    NoError: null
    SysFatal: null
    SysTransient: null
    DestinationInvalid: null
    CanisterReject: null
    CanisterError: null
    Unknown: null


# TODO we might want this to act more like CallResult
class NotifyResult(Variant, total=False):
    Ok: null
    Err: RejectionCode


class StableMemoryError(Variant, total=False):
    OutOfMemory: null
    OutOfBounds: null


class StableGrowResult(Variant, total=False):
    Ok: nat32
    Err: StableMemoryError


class Stable64GrowResult(Variant, total=False):
    Ok: nat64
    Err: StableMemoryError


FuncTuple = tuple[Principal, str]


def Func(callable: Callable[..., Any]) -> type[FuncTuple]:
    return FuncTuple


def get_first_called_function_name() -> str:
    first_frame = get_first_frame(sys._getframe())  # type: ignore
    return first_frame.f_code.co_name


def get_first_frame(current_frame: Any) -> Any:
    previous_frame = current_frame.f_back

    if previous_frame is None:
        return current_frame

    return get_first_frame(previous_frame)


class ic(Generic[T]):
    @staticmethod
    def accept_message():
        _basilisk_ic.accept_message()  # type: ignore

    @staticmethod
    def arg_data_raw() -> blob:
        return _basilisk_ic.arg_data_raw()  # type: ignore

    @staticmethod
    def arg_data_raw_size() -> nat32:
        return _basilisk_ic.arg_data_raw_size()  # type: ignore

    @staticmethod
    def call_raw(
        canister_id: Principal, method: str, args_raw: blob, payment: nat64
    ) -> CallResult[T]:
        return AsyncInfo(
            "call_raw", [canister_id, method, args_raw, payment]
        )  # type: ignore

    @staticmethod
    def call_raw128(
        canister_id: Principal, method: str, args_raw: blob, payment: nat
    ) -> CallResult[T]:
        return AsyncInfo(
            "call_raw128", [canister_id, method, args_raw, payment]
        )  # type: ignore

    @staticmethod
    def caller() -> Principal:
        """Returns the caller of the current call.

        Returns:
            (Principal): the caller of the current call.
        Raises:
            TypeError: the caller could not be converted to a Principal.
        """
        return _basilisk_ic.caller()  # type: ignore

    @staticmethod
    def candid_encode(candid_string: str) -> blob:
        """Converts the provided string into a Candid value.

        Args:
            candid_string (str): a string representation of a Candid value

        Returns:
            (blob): a Candid value

        Raises:
            CandidError: an error occurred while processing the input.
            TypeError: the provided value was not of the correct type.
        """
        return _basilisk_ic.candid_encode(candid_string)  # type: ignore

    @staticmethod
    def candid_decode(candid_encoded: blob) -> str:
        """Converts the provided Candid bytes into a string representation.

        Args:
            candid_encode (blob): a blob representing a Candid value.

        Returns:
            (blob): a string representation of the value.

        Raises:
            CandidError: an error occurred while processing the input.
            TypeError: the provided value was not of the correct type.
        """
        return _basilisk_ic.candid_decode(candid_encoded)  # type: ignore

    @staticmethod
    def canister_balance() -> nat64:
        return _basilisk_ic.canister_balance()  # type: ignore

    @staticmethod
    def canister_balance128() -> nat:
        return _basilisk_ic.canister_balance128()  # type: ignore

    @staticmethod
    def clear_timer(id: TimerId) -> None:
        return _basilisk_ic.clear_timer(id)  # type: ignore

    @staticmethod
    def data_certificate() -> Opt[blob]:
        return _basilisk_ic.data_certificate()  # type: ignore

    @staticmethod
    def id() -> Principal:
        return _basilisk_ic.id()  # type:ignore

    @staticmethod
    def method_name() -> str:
        return _basilisk_ic.method_name()  # type:ignore

    @staticmethod
    def msg_cycles_accept(max_amount: nat64) -> nat64:
        return _basilisk_ic.msg_cycles_accept(max_amount)  # type: ignore

    @staticmethod
    def msg_cycles_accept128(max_amount: nat) -> nat:
        return _basilisk_ic.msg_cycles_accept128(max_amount)  # type: ignore

    @staticmethod
    def msg_cycles_available() -> nat64:
        return _basilisk_ic.msg_cycles_available()  # type: ignore

    @staticmethod
    def msg_cycles_available128() -> nat:
        return _basilisk_ic.msg_cycles_available128()  # type: ignore

    @staticmethod
    def msg_cycles_refunded() -> nat64:
        return _basilisk_ic.msg_cycles_refunded()  # type: ignore

    @staticmethod
    def msg_cycles_refunded128() -> nat:
        return _basilisk_ic.msg_cycles_refunded128()  # type: ignore

    @staticmethod
    def notify_raw(
        canister_id: Principal, method: str, args_raw: blob, payment: nat
    ) -> NotifyResult:
        return _basilisk_ic.notify_raw(  # type: ignore
            canister_id, method, args_raw, payment
        )

    @staticmethod
    def performance_counter(counter_type: nat32) -> nat64:
        return _basilisk_ic.performance_counter(counter_type)  # type: ignore

    @staticmethod
    def print(*args: Any):
        string_list = [str(arg) for arg in args]
        _basilisk_ic.print(" ".join(string_list))  # type: ignore

    @staticmethod
    def reject(x: str):
        _basilisk_ic.reject(x)  # type: ignore

    @staticmethod
    def reject_code() -> RejectionCode:
        return _basilisk_ic.reject_code()  # type: ignore

    @staticmethod
    def reject_message() -> str:
        return _basilisk_ic.reject_message()  # type: ignore

    @staticmethod
    def reply(value: Any):
        first_called_function_name = get_first_called_function_name()
        (_basilisk_ic.reply(first_called_function_name, value))  # type: ignore

    @staticmethod
    def reply_raw(x: Any):
        _basilisk_ic.reply_raw(x)  # type: ignore

    @staticmethod
    def set_certified_data(data: blob):
        _basilisk_ic.set_certified_data(data)  # type: ignore

    @staticmethod
    def set_timer(delay: Duration, func: Callable[[], Any]) -> TimerId:
        return _basilisk_ic.set_timer(delay, func)  # type: ignore

    @staticmethod
    def set_timer_interval(interval: Duration, func: Callable[[], Any]) -> TimerId:
        return _basilisk_ic.set_timer_interval(interval, func)  # type: ignore

    @staticmethod
    def stable_bytes() -> blob:
        return _basilisk_ic.stable_bytes()  # type: ignore

    @staticmethod
    def stable_grow(new_pages: nat32) -> StableGrowResult:
        return _basilisk_ic.stable_grow(new_pages)  # type: ignore

    @staticmethod
    def stable_read(offset: nat32, length: nat32) -> blob:
        return _basilisk_ic.stable_read(offset, length)  # type: ignore

    @staticmethod
    def stable_size() -> nat32:
        return _basilisk_ic.stable_size()  # type: ignore

    @staticmethod
    def stable_write(offset: nat32, buf: blob):
        _basilisk_ic.stable_write(offset, buf)  # type: ignore

    @staticmethod
    def stable64_grow(new_pages: nat64) -> Stable64GrowResult:
        return _basilisk_ic.stable64_grow(new_pages)  # type: ignore

    @staticmethod
    def stable64_read(offset: nat64, length: nat64) -> blob:
        return _basilisk_ic.stable64_read(offset, length)  # type: ignore

    @staticmethod
    def stable64_size() -> nat64:
        return _basilisk_ic.stable64_size()  # type: ignore

    @staticmethod
    def stable64_write(offset: nat64, buf: blob):
        _basilisk_ic.stable64_write(offset, buf)  # type: ignore

    @staticmethod
    def time() -> nat64:
        return _basilisk_ic.time()  # type: ignore

    @staticmethod
    def trap(message: str) -> NoReturn:  # type: ignore
        _basilisk_ic.trap(message)  # type: ignore


class Service:
    canister_id: Principal

    def __init__(self, canister_id: Principal):
        self.canister_id = canister_id


P = ParamSpec("P")


class AsyncInfo:
    name: str
    args: list[Any]

    def __init__(self, name: str, args: list[Any]):
        self.name = name
        self.args = args

    def with_cycles(self, cycles: nat64) -> "AsyncInfo":
        return AsyncInfo("call_with_payment", [*self.args, cycles])

    def with_cycles128(self, cycles: nat) -> "AsyncInfo":
        return AsyncInfo("call_with_payment128", [*self.args, cycles])

    def notify(self) -> NotifyResult:
        qualname: str = self.args[1]
        with_payment = (
            "with_payment128_"
            if self.name == "call_with_payment" or self.name == "call_with_payment128"
            else ""
        )
        notify_function_name = (
            f'notify_{with_payment}{qualname.replace(".", "_")}_wrapper'
        )

        return getattr(_basilisk_ic, notify_function_name)(self.args)  # type: ignore


# TODO this decorator is removing the static type checking of the self parameter for instance methods
# TODO watch out for *kwargs
def service_method(func: Callable[P, T]) -> Callable[P, CallResult[T]]:
    def intermediate_func(*args):  # type: ignore
        the_self = args[0]  # type: ignore
        selfless_args = args[1:]  # type: ignore

        return AsyncInfo(
            "call",
            [the_self.canister_id, func.__qualname__, *selfless_args],  # type: ignore
        )

    return intermediate_func  # type: ignore


def service_query(func: Callable[P, T]) -> Callable[P, CallResult[T]]:
    return service_method(func)


def service_update(func: Callable[P, T]) -> Callable[P, CallResult[T]]:
    return service_method(func)


K = TypeVar("K")
V = TypeVar("V")


class KeyTooLarge(Record):
    given: nat32
    max: nat32


class ValueTooLarge(Record):
    given: nat32
    max: nat32


class StableBTreeMap(Generic[K, V]):
    """
    A map based on a self-balancing tree that persists across canister upgrades.
    """

    def __init__(self, memory_id: nat8, max_key_size: int, max_value_size: int):
        """
        Initialize the stable B-tree map.

        :param memory_id: The memory ID of the stable B-tree map.
        :param max_key_size: The maximum size of the keys (in bytes) in the map.
        :param max_value_size: The maximum size of the values (in bytes) in the map.
        """
        self.memory_id = memory_id

    def contains_key(self, key: K) -> bool:
        """
        Check if the map contains a key.

        :param key: The key to check for in the map.
        :return: True if the key is in the map, False otherwise.
        """
        return _basilisk_ic.stable_b_tree_map_contains_key(self.memory_id, key)  # type: ignore

    def get(self, key: K) -> Opt[V]:
        """
        Get the value associated with a key in the map.

        :param key: The key to get the value for.
        :return: The value associated with the key, or None if the key is not in the map.
        """
        return _basilisk_ic.stable_b_tree_map_get(self.memory_id, key)  # type: ignore

    def insert(self, key: K, value: V) -> Opt[V]:
        """
        Insert a key-value pair into the map.

        :param key: The key to insert.
        :param value: The value to insert.
        :return: An instance of InsertResult containing an ok attribute if the insertion succeeded
        or an err attribute if the insertion failed. If the insertion succeeded the ok attribute
        will contain the previous value associated with the key, if any. If the insertion failed,
        the err attribute will contain an instance of InsertError indicating the reason for the
        failure.
        """
        return _basilisk_ic.stable_b_tree_map_insert(self.memory_id, key, value)  # type: ignore

    def is_empty(self) -> bool:
        """
        Check if the map is empty.

        :return: True if the map is empty, False otherwise.
        """
        return _basilisk_ic.stable_b_tree_map_is_empty(self.memory_id)  # type: ignore

    def items(self) -> Vec[Tuple[K, V]]:
        """
        Get a list of all key-value pairs in the map.

        :return: A list of tuples containing all key-value pairs in the map.
        """
        return _basilisk_ic.stable_b_tree_map_items(self.memory_id)  # type: ignore

    def keys(self) -> Vec[K]:
        """
        Get a list of all keys in the map.

        :return: A list of all keys in the map.
        """
        return _basilisk_ic.stable_b_tree_map_keys(self.memory_id)  # type: ignore

    def len(self) -> nat64:
        """
        Get the number of key-value pairs in the map.

        :return: The number of key-value pairs in the map.
        """
        return _basilisk_ic.stable_b_tree_map_len(self.memory_id)  # type: ignore

    def remove(self, key: K) -> Opt[V]:
        """
        Remove a key-value pair from the map.

        :param key: The key of the key-value pair to remove.
        :return: The value associated with the key, or None if the key is not in the map.
        """
        return _basilisk_ic.stable_b_tree_map_remove(self.memory_id, key)  # type: ignore

    def values(self) -> Vec[V]:
        """
        Get a list of all values in the map.

        :return: A list of all values in the map.
        """
        return _basilisk_ic.stable_b_tree_map_values(self.memory_id)  # type: ignore


def match(
    variant: Union[TypedDict, object], matcher: dict[str, Callable[[Any], T]]
) -> T:
    if isinstance(variant, dict):
        for key, value in matcher.items():
            if key in variant:
                return value(variant[key])

            if key == "_":
                return value(None)
    else:
        # This only works for Result (Ok, Err) objects
        # This is temporary until we either use dataclasses for all variants or TypedDicts for all variants
        err_value = getattr(variant, "Err", None)

        if err_value is not None:
            return matcher["Err"](err_value)

        return matcher["Ok"](getattr(variant, "Ok"))

    raise Exception("No matching case found")


# region Exceptions
class Error(Exception):
    """Base exception for all errors raised by Basilisk"""

    pass


class CandidError(Error):
    """Raised when converting to/from Candid values."""

    pass


# endregion Exceptions



import os
import re
import shutil
import subprocess
import sys
from typing import Callable

import basilisk
from basilisk.colors import red
from basilisk.types import Paths


def run_basilisk_generate_or_exit(paths: Paths, cargo_env: dict[str, str], verbose: bool):
    bin_path, bin_path_debug = construct_bin_paths(paths)
    should_rebuild = check_rebuild(bin_path)
    if should_rebuild:
        run_build(bin_path, bin_path_debug, paths, cargo_env, verbose)
    run_basilisk_generate(bin_path, paths, cargo_env, verbose)


def construct_bin_paths(paths: Paths) -> tuple[str, str]:
    bin_path = (
        f"{paths['global_basilisk_config_dir']}/{basilisk.__version__}/bin/basilisk_generate"
    )
    bin_path_debug = f"{paths['global_basilisk_target_dir']}/debug/basilisk_generate"
    return bin_path, bin_path_debug


def check_rebuild(bin_path: str) -> bool:
    return not os.path.exists(bin_path) or os.environ.get("BASILISK_REBUILD") == "true"


def run_build(
    bin_path: str,
    bin_path_debug: str,
    paths: Paths,
    cargo_env: dict[str, str],
    verbose: bool,
):
    build_result = execute_build_process(paths, cargo_env, verbose)
    validate_process(build_result, parse_basilisk_generate_error)
    copy_file(bin_path_debug, bin_path)


def execute_build_process(
    paths: Paths, cargo_env: dict[str, str], verbose: bool
) -> subprocess.CompletedProcess[bytes]:
    return subprocess.run(
        [
            f"{paths['global_basilisk_rust_bin_dir']}/cargo",
            "build",
            f"--manifest-path={paths['canister']}/basilisk_generate/Cargo.toml",
        ],
        capture_output=not verbose,
        env=cargo_env,
    )


def copy_file(source: str, destination: str):
    shutil.copy(source, destination)


def run_basilisk_generate(
    bin_path: str, paths: Paths, cargo_env: dict[str, str], verbose: bool
):
    generate_result = execute_generate_process(bin_path, paths, cargo_env, verbose)
    validate_process(generate_result, parse_basilisk_generate_error)


def execute_generate_process(
    bin_path: str, paths: Paths, cargo_env: dict[str, str], verbose: bool
) -> subprocess.CompletedProcess[bytes]:
    return subprocess.run(
        [
            bin_path,
            paths["py_file_names_file"],
            paths["py_entry_module_name"],
            paths["lib"],
            basilisk.__version__,
        ],
        capture_output=not verbose,
        env=cargo_env,
    )


def validate_process(
    result: subprocess.CompletedProcess[bytes], error_parser: Callable[[bytes], str]
) -> None:
    if result.returncode != 0:
        print_error_and_exit_basilisk_generate(result, error_parser)


def print_error_and_exit_basilisk_generate(
    result: subprocess.CompletedProcess[bytes], error_parser: Callable[[bytes], str]
) -> None:
    print(red("\n💣 Basilisk error: compilation\n"))
    print(error_parser(result.stderr))
    print_help_message()
    sys.exit(1)


def print_help_message():
    print(
        "\nFor help reach out in the #python channel of the ICP Developer Community discord:"
    )
    print("\nhttps://discord.com/channels/748416164832608337/1019372359775440988\n")
    print("💀 Build failed")


def parse_basilisk_generate_error(stdout: bytes) -> str:
    err = stdout.decode("utf-8")
    std_err_lines = err.splitlines()
    try:
        line_where_error_message_starts = next(
            i
            for i, v in enumerate(std_err_lines)
            if v.startswith("thread 'main' panicked at '")
        )
        line_where_error_message_ends = next(
            i for i, v in enumerate(std_err_lines) if "', src/" in v
        )
    except:
        return err

    err_lines = std_err_lines[
        line_where_error_message_starts : line_where_error_message_ends + 1
    ]
    err_lines[0] = err_lines[0].replace("thread 'main' panicked at '", "")
    err_lines[-1] = re.sub("', src/.*", "", err_lines[-1])

    return red("\n".join(err_lines))



import sys as _sys

# --- Bootstrap import system for WASI CPython ---
# When _Py_InitializeMain is skipped, sys.meta_path may not be consulted
# and many stdlib modules aren't available (no filesystem).
# Strategy: wrap builtins.__import__ to catch ModuleNotFoundError and
# auto-create empty stub modules.  This is more reliable than sys.meta_path
# which may not be fully functional without _Py_InitializeMain.

# Also try to install BuiltinImporter/FrozenImporter if missing
_frozen = _sys.modules.get('_frozen_importlib')
if _frozen:
    if not _sys.meta_path:
        _sys.meta_path.append(_frozen.BuiltinImporter)
        _sys.meta_path.append(_frozen.FrozenImporter)
try:
    del _frozen
except NameError:
    pass

import builtins as _builtins
_orig_import = _builtins.__import__
_ModuleType = type(_sys)

def _wasi_safe_import(name, globals=None, locals=None, fromlist=(), level=0):
    try:
        return _orig_import(name, globals, locals, fromlist, level)
    except (ModuleNotFoundError, ImportError):
        # For relative imports (level > 0), don't stub - let it fail normally
        if level > 0:
            raise
        # Only stub truly missing modules.
        # If the module is already in sys.modules, the error came from INSIDE
        # module loading (e.g. a sub-import failed) - let it propagate.
        if name in _sys.modules:
            raise
        # Create a stub module so the import doesn't crash
        mod = _ModuleType(name)
        mod.__file__ = "<wasi-stub>"
        mod.__path__ = []
        mod.__package__ = name
        _sys.modules[name] = mod
        return mod

# NOTE: _wasi_safe_import is installed AFTER all rich stdlib stubs below.
# This ensures try/except import blocks use _orig_import and our rich stubs
# get registered, not empty wasi-stubs.

# --- Patch basilisk module with missing classes ---
# The CPython template WASM shim may not include StableBTreeMap, Service, match.
# Add them here so user code that imports them from basilisk works.
_bmod = _sys.modules.get('basilisk')
if _bmod and not hasattr(_bmod, 'StableBTreeMap'):
    try:
        import _basilisk_ic as _bic
    except (ImportError, ModuleNotFoundError):
        _bic = None

    class _StableBTreeMap:
        def __init__(self, memory_id, max_key_size=0, max_value_size=0):
            self.memory_id = memory_id
        def __class_getitem__(cls, params):
            return cls
        def _fn(self, op):
            return getattr(_bic, f"stable_b_tree_map_{self.memory_id}_{op}")
        def contains_key(self, key):
            return self._fn("contains_key")(key)
        def get(self, key):
            return self._fn("get")(key)
        def insert(self, key, value):
            return self._fn("insert")(key, value)
        def is_empty(self):
            return self._fn("is_empty")()
        def items(self):
            return self._fn("items")()
        def keys(self):
            return self._fn("keys")()
        def len(self):
            return self._fn("len")()
        def remove(self, key):
            return self._fn("remove")(key)
        def values(self):
            return self._fn("values")()
    _bmod.StableBTreeMap = _StableBTreeMap

    class _Service:
        def __init__(self, canister_id):
            self.canister_id = canister_id
    _bmod.Service = _Service

    def _match(variant, matcher):
        if isinstance(variant, dict):
            for key, value in matcher.items():
                if key in variant:
                    return value(variant[key])
                if key == "_":
                    return value(None)
        else:
            err_value = getattr(variant, "Err", None)
            if err_value is not None:
                return matcher["Err"](err_value)
            return matcher["Ok"](getattr(variant, "Ok"))
        raise Exception("No matching case found")
    _bmod.match = _match

    try:
        del _bic, _StableBTreeMap, _Service, _match
    except NameError:
        pass
try:
    del _bmod
except NameError:
    pass

# --- frozen stdlib: json module (pure Python, no C extensions) ---
# On WASI/IC there is no filesystem, so stdlib packages like `json`
# aren't importable. This registers a minimal pure-Python json module
# in sys.modules so that `import json` works out of the box.

def _register_json():
    _ESCAPE = {'"': '\\"', '\\': '\\\\', '\b': '\\b', '\f': '\\f',
               '\n': '\\n', '\r': '\\r', '\t': '\\t'}

    def _enc(s):
        parts = ['"']
        for c in s:
            if c in _ESCAPE:
                parts.append(_ESCAPE[c])
            elif ord(c) < 0x20:
                parts.append(f'\\u{ord(c):04x}')
            else:
                parts.append(c)
        parts.append('"')
        return ''.join(parts)

    class JSONDecodeError(ValueError):
        pass

    def dumps(obj, ensure_ascii=True, sort_keys=False, indent=None,
              separators=None, default=None, **kw):
        def _e(o):
            if o is None: return "null"
            if o is True: return "true"
            if o is False: return "false"
            if isinstance(o, str): return _enc(o)
            if isinstance(o, int): return str(o)
            if isinstance(o, float):
                if o != o: return "NaN"
                if o == float("inf"): return "Infinity"
                if o == float("-inf"): return "-Infinity"
                return repr(o)
            if isinstance(o, (list, tuple)):
                return "[" + ",".join(_e(v) for v in o) + "]"
            if isinstance(o, dict):
                items = sorted(o.items()) if sort_keys else o.items()
                return "{" + ",".join(_enc(str(k)) + ":" + _e(v)
                                      for k, v in items) + "}"
            if default is not None: return _e(default(o))
            raise TypeError(
                f"Object of type {type(o).__name__} is not JSON serializable")
        return _e(obj)

    def _parse_string(s, idx, n):
        parts = []
        while idx < n:
            c = s[idx]
            if c == '"':
                return ''.join(parts), idx + 1
            if c == '\\':
                idx += 1
                if idx >= n:
                    raise JSONDecodeError("Unterminated string escape")
                esc = s[idx]
                if esc == '"': parts.append('"')
                elif esc == '\\': parts.append('\\')
                elif esc == '/': parts.append('/')
                elif esc == 'b': parts.append('\b')
                elif esc == 'f': parts.append('\f')
                elif esc == 'n': parts.append('\n')
                elif esc == 'r': parts.append('\r')
                elif esc == 't': parts.append('\t')
                elif esc == 'u':
                    h = s[idx + 1:idx + 5]
                    if len(h) < 4:
                        raise JSONDecodeError("Truncated \\uXXXX escape")
                    parts.append(chr(int(h, 16)))
                    idx += 4
                else:
                    parts.append(esc)
                idx += 1
            else:
                parts.append(c)
                idx += 1
        raise JSONDecodeError("Unterminated string")

    def loads(s, **kw):
        if isinstance(s, (bytes, bytearray)):
            s = s.decode("utf-8")
        n = len(s)
        idx = [0]

        def _ws():
            while idx[0] < n and s[idx[0]] in " \t\n\r":
                idx[0] += 1

        def _val():
            _ws()
            if idx[0] >= n:
                raise JSONDecodeError("Unexpected end of JSON")
            c = s[idx[0]]
            if c == '"':
                r, new_idx = _parse_string(s, idx[0] + 1, n)
                idx[0] = new_idx
                return r
            if c == '{':
                idx[0] += 1; _ws(); d = {}
                if idx[0] < n and s[idx[0]] == '}':
                    idx[0] += 1; return d
                while True:
                    _ws(); k = _val(); _ws()
                    if idx[0] >= n or s[idx[0]] != ':':
                        raise JSONDecodeError("Expected ':'")
                    idx[0] += 1; v = _val(); d[k] = v; _ws()
                    if idx[0] < n and s[idx[0]] == ',':
                        idx[0] += 1
                    elif idx[0] < n and s[idx[0]] == '}':
                        idx[0] += 1; return d
                    else:
                        raise JSONDecodeError("Expected ',' or '}'")
            if c == '[':
                idx[0] += 1; _ws(); a = []
                if idx[0] < n and s[idx[0]] == ']':
                    idx[0] += 1; return a
                while True:
                    a.append(_val()); _ws()
                    if idx[0] < n and s[idx[0]] == ',':
                        idx[0] += 1
                    elif idx[0] < n and s[idx[0]] == ']':
                        idx[0] += 1; return a
                    else:
                        raise JSONDecodeError("Expected ',' or ']'")
            if s[idx[0]:idx[0] + 4] == "true":
                idx[0] += 4; return True
            if s[idx[0]:idx[0] + 5] == "false":
                idx[0] += 5; return False
            if s[idx[0]:idx[0] + 4] == "null":
                idx[0] += 4; return None
            # Number
            st = idx[0]
            if idx[0] < n and s[idx[0]] == '-':
                idx[0] += 1
            while idx[0] < n and s[idx[0]].isdigit():
                idx[0] += 1
            flt = False
            if idx[0] < n and s[idx[0]] == '.':
                flt = True; idx[0] += 1
                while idx[0] < n and s[idx[0]].isdigit():
                    idx[0] += 1
            if idx[0] < n and s[idx[0]] in "eE":
                flt = True; idx[0] += 1
                if idx[0] < n and s[idx[0]] in "+-":
                    idx[0] += 1
                while idx[0] < n and s[idx[0]].isdigit():
                    idx[0] += 1
            ns = s[st:idx[0]]
            if not ns:
                raise JSONDecodeError(f"Unexpected char at {idx[0]}")
            return float(ns) if flt else int(ns)

        return _val()

    m = type(_sys)("json")
    m.__file__ = "<frozen json>"
    m.dumps = dumps
    m.loads = loads
    m.JSONDecodeError = JSONDecodeError
    _sys.modules["json"] = m

try:
    import json
except ImportError:
    _register_json()
del _register_json


# --- frozen stdlib: random module (pure Python, no C extensions) ---
# The CPython canister template seeds Python's random module with IC consensus
# randomness (raw_rand) at init. On WASI there is no filesystem so stdlib
# `random` isn't importable. This provides a minimal pure-Python implementation.

def _register_random():
    # Mersenne Twister constants
    _N = 624
    _M = 397
    _MATRIX_A = 0x9908b0df
    _UPPER_MASK = 0x80000000
    _LOWER_MASK = 0x7fffffff

    class Random:
        def __init__(self, x=None):
            self._mt = [0] * _N
            self._mti = _N + 1
            if x is not None:
                self.seed(x)
            else:
                self.seed(0)

        def seed(self, a=None):
            if a is None:
                a = 0
            if isinstance(a, (bytes, bytearray)):
                # Convert bytes to integer (big-endian)
                a = int.from_bytes(a, 'big')
            if isinstance(a, float):
                a = int(a)
            a = abs(a)
            self._mt[0] = a & 0xffffffff
            for i in range(1, _N):
                self._mt[i] = (1812433253 * (self._mt[i - 1] ^ (self._mt[i - 1] >> 30)) + i) & 0xffffffff
            self._mti = _N

        def _generate(self):
            mag01 = [0, _MATRIX_A]
            for kk in range(_N - _M):
                y = (self._mt[kk] & _UPPER_MASK) | (self._mt[kk + 1] & _LOWER_MASK)
                self._mt[kk] = self._mt[kk + _M] ^ (y >> 1) ^ mag01[y & 1]
            for kk in range(_N - _M, _N - 1):
                y = (self._mt[kk] & _UPPER_MASK) | (self._mt[kk + 1] & _LOWER_MASK)
                self._mt[kk] = self._mt[kk + (_M - _N)] ^ (y >> 1) ^ mag01[y & 1]
            y = (self._mt[_N - 1] & _UPPER_MASK) | (self._mt[0] & _LOWER_MASK)
            self._mt[_N - 1] = self._mt[_M - 1] ^ (y >> 1) ^ mag01[y & 1]
            self._mti = 0

        def _genrand_int32(self):
            if self._mti >= _N:
                self._generate()
            y = self._mt[self._mti]
            self._mti += 1
            y ^= (y >> 11)
            y ^= (y << 7) & 0x9d2c5680
            y ^= (y << 15) & 0xefc60000
            y ^= (y >> 18)
            return y

        def random(self):
            a = self._genrand_int32() >> 5
            b = self._genrand_int32() >> 6
            return (a * 67108864.0 + b) * (1.0 / 9007199254740992.0)

        def randint(self, a, b):
            return a + int(self.random() * (b - a + 1))

        def randrange(self, start, stop=None, step=1):
            if stop is None:
                return int(self.random() * start)
            return start + step * int(self.random() * ((stop - start + step - 1) // step))

        def choice(self, seq):
            return seq[int(self.random() * len(seq))]

        def shuffle(self, x):
            for i in range(len(x) - 1, 0, -1):
                j = int(self.random() * (i + 1))
                x[i], x[j] = x[j], x[i]

        def sample(self, population, k):
            pool = list(population)
            n = len(pool)
            result = []
            for i in range(k):
                j = int(self.random() * (n - i))
                result.append(pool[j])
                pool[j] = pool[n - i - 1]
            return result

        def uniform(self, a, b):
            return a + (b - a) * self.random()

        def getrandbits(self, k):
            if k <= 0:
                raise ValueError("number of bits must be greater than zero")
            numbytes = (k + 7) // 8
            x = int.from_bytes(bytes([self._genrand_int32() & 0xff for _ in range(numbytes)]), 'big')
            return x >> (numbytes * 8 - k)

    _inst = Random()

    m = type(_sys)("random")
    m.__file__ = "<frozen random>"
    m.Random = Random
    m.seed = _inst.seed
    m.random = _inst.random
    m.randint = _inst.randint
    m.randrange = _inst.randrange
    m.choice = _inst.choice
    m.shuffle = _inst.shuffle
    m.sample = _inst.sample
    m.uniform = _inst.uniform
    m.getrandbits = _inst.getrandbits
    _sys.modules["random"] = m

try:
    import random
except ImportError:
    _register_random()
del _register_random


# --- frozen stdlib: time module ---
def _register_time():
    def _time():
        return 0.0
    def _sleep(secs):
        pass
    def _monotonic():
        return 0.0
    m = type(_sys)("time")
    m.__file__ = "<frozen time>"
    m.time = _time
    m.sleep = _sleep
    m.monotonic = _monotonic
    m.perf_counter = _monotonic
    m.time_ns = lambda: 0
    _sys.modules["time"] = m

try:
    import time
except ImportError:
    _register_time()
del _register_time


# --- frozen stdlib: datetime module ---
def _register_datetime():
    class timedelta:
        def __init__(self, days=0, seconds=0, microseconds=0,
                     milliseconds=0, minutes=0, hours=0, weeks=0):
            total = (days * 86400 + hours * 3600 + minutes * 60 + seconds
                     + milliseconds / 1000 + microseconds / 1000000 + weeks * 604800)
            self._total_seconds = total
        def total_seconds(self):
            return self._total_seconds
        def __repr__(self):
            return f"timedelta(seconds={self._total_seconds})"

    class datetime:
        def __init__(self, year=1970, month=1, day=1, hour=0, minute=0,
                     second=0, microsecond=0):
            self.year = year; self.month = month; self.day = day
            self.hour = hour; self.minute = minute; self.second = second
            self.microsecond = microsecond
        @staticmethod
        def now():
            return datetime()
        @staticmethod
        def fromtimestamp(ts):
            # Minimal: just store ts, return stub
            d = datetime()
            d._ts = ts
            return d
        def strftime(self, fmt):
            return f"{self.year:04d}-{self.month:02d}-{self.day:02d} {self.hour:02d}:{self.minute:02d}:{self.second:02d}.{self.microsecond:06d}"
        def __repr__(self):
            return self.strftime("%Y-%m-%d %H:%M:%S.%f")

    m = type(_sys)("datetime")
    m.__file__ = "<frozen datetime>"
    m.datetime = datetime
    m.timedelta = timedelta
    _sys.modules["datetime"] = m

try:
    import datetime
except ImportError:
    _register_datetime()
del _register_datetime


# --- frozen stdlib: typing module ---
def _register_typing():
    m = type(_sys)("typing")
    m.__file__ = "<frozen typing>"
    # All typing constructs are no-ops at runtime
    m.Any = object
    m.Union = None
    m.Optional = lambda x: x
    m.List = list
    m.Dict = dict
    m.Set = set
    m.Tuple = tuple
    m.Type = type
    m.Callable = object
    m.Iterator = object
    m.Generator = object
    m.Iterable = object
    m.Sequence = object
    m.Mapping = object
    m.MutableMapping = object
    m.Deque = list
    m.FrozenSet = frozenset
    m.Counter = dict
    m.OrderedDict = dict
    m.DefaultDict = dict
    m.NamedTuple = tuple
    m.IO = object
    m.TextIO = object
    m.BinaryIO = object
    m.Pattern = object
    m.Match = object
    m.AnyStr = object
    m.SupportsInt = object
    m.SupportsFloat = object
    m.SupportsComplex = object
    m.SupportsBytes = object
    m.SupportsAbs = object
    m.SupportsRound = object
    m.ClassVar = None
    m.Final = None
    m.Literal = None
    m.Annotated = None
    m.TypeAlias = None
    m.NoReturn = None
    m.TYPE_CHECKING = False
    def _identity(x=None): return x
    m.TypeVar = lambda name, *a, **kw: object
    m.ParamSpec = lambda name, *a, **kw: object
    m.Generic = object
    m.Protocol = object
    m.TypedDict = type
    m.overload = lambda f: f
    m.cast = lambda t, v: v
    m.no_type_check = lambda f: f
    m.runtime_checkable = lambda cls: cls
    _sys.modules["typing"] = m

try:
    import typing
except ImportError:
    _register_typing()
del _register_typing


# --- frozen stdlib: abc module ---
def _register_abc():
    def abstractmethod(f):
        return f
    class ABCMeta(type):
        pass
    class ABC(metaclass=ABCMeta):
        pass
    m = type(_sys)("abc")
    m.__file__ = "<frozen abc>"
    m.ABC = ABC
    m.ABCMeta = ABCMeta
    m.abstractmethod = abstractmethod
    _sys.modules["abc"] = m

try:
    import abc
except ImportError:
    _register_abc()
del _register_abc


# --- frozen stdlib: weakref module ---
def _register_weakref():
    class WeakValueDictionary(dict):
        """Stub: acts as a regular dict (no weak references in WASI)."""
        pass
    class WeakSet(set):
        pass
    def ref(obj, callback=None):
        return lambda: obj
    m = type(_sys)("weakref")
    m.__file__ = "<frozen weakref>"
    m.WeakValueDictionary = WeakValueDictionary
    m.WeakSet = WeakSet
    m.ref = ref
    _sys.modules["weakref"] = m

try:
    import weakref
except ImportError:
    _register_weakref()
del _register_weakref


# --- frozen stdlib: enum module ---
def _register_enum():
    class EnumMeta(type):
        def __new__(mcs, name, bases, namespace):
            cls = super().__new__(mcs, name, bases, namespace)
            return cls
        def __iter__(cls):
            return iter([])
    class Enum(metaclass=EnumMeta):
        pass
    class IntEnum(int, Enum):
        pass
    m = type(_sys)("enum")
    m.__file__ = "<frozen enum>"
    m.Enum = Enum
    m.IntEnum = IntEnum
    m.EnumMeta = EnumMeta
    m.unique = lambda cls: cls
    m.auto = lambda: 0
    _sys.modules["enum"] = m

try:
    import enum
except ImportError:
    _register_enum()
del _register_enum


# --- frozen stdlib: collections module ---
def _register_collections():
    m = type(_sys)("collections")
    m.__file__ = "<frozen collections>"
    m.deque = list
    m.OrderedDict = dict
    m.defaultdict = dict
    m.Counter = dict
    m.namedtuple = lambda name, fields: type(name, (tuple,), {})
    m.ChainMap = dict
    _sys.modules["collections"] = m

try:
    import collections
    collections.deque  # verify it's real
except (ImportError, AttributeError):
    _register_collections()
del _register_collections


# --- frozen stdlib: dataclasses module ---
def _register_dataclasses():
    def dataclass(cls=None, **kw):
        if cls is None:
            return lambda c: c
        return cls
    def field(**kw):
        return kw.get('default', kw.get('default_factory', lambda: None)())
    m = type(_sys)("dataclasses")
    m.__file__ = "<frozen dataclasses>"
    m.dataclass = dataclass
    m.field = field
    m.fields = lambda cls: []
    m.asdict = lambda obj: {}
    m.astuple = lambda obj: ()
    m.replace = lambda obj, **kw: obj
    m.is_dataclass = lambda obj: False
    m.FrozenInstanceError = type('FrozenInstanceError', (AttributeError,), {})
    _sys.modules["dataclasses"] = m

try:
    import dataclasses
    dataclasses.dataclass  # verify it's real
except (ImportError, AttributeError):
    _register_dataclasses()
del _register_dataclasses


# --- frozen stdlib: functools module ---
def _register_functools():
    def lru_cache(maxsize=128, typed=False):
        def decorator(f):
            return f
        if callable(maxsize):
            return maxsize
        return decorator
    m = type(_sys)("functools")
    m.__file__ = "<frozen functools>"
    m.lru_cache = lru_cache
    m.wraps = lambda f: lambda g: g
    m.partial = lambda f, *a, **kw: lambda *a2, **kw2: f(*a, *a2, **{**kw, **kw2})
    m.reduce = lambda f, it, *init: None
    m.cache = lru_cache
    m.cached_property = property
    m.total_ordering = lambda cls: cls
    m.singledispatch = lambda f: f
    _sys.modules["functools"] = m

try:
    import functools
    functools.wraps  # verify it's real
except (ImportError, AttributeError):
    _register_functools()
del _register_functools

# --- Install the universal fallback import wrapper ---
# This MUST be after all rich stdlib stubs above, so that try/except import
# blocks use _orig_import and the rich stubs get registered properly.
_builtins.__import__ = _wasi_safe_import



"""
Basilisk CLI — scaffold and build IC canister projects in Python.

Usage:
    basilisk new [--backend cpython|rustpython] <project_name>
    basilisk build                 Build the canister in the current directory
    basilisk --version             Print version
"""

import os
import sys
from pathlib import Path


def cmd_new(project_name: str, backend: str = "cpython"):
    """Scaffold a new basilisk canister project."""
    project_dir = Path(project_name)

    if project_dir.exists():
        print(f"Error: directory '{project_name}' already exists.", file=sys.stderr)
        sys.exit(1)

    # Validate project name (must be a valid canister name)
    if not project_name.replace("_", "").replace("-", "").isalnum():
        print(f"Error: '{project_name}' is not a valid project name. Use alphanumeric, dashes, and underscores.", file=sys.stderr)
        sys.exit(1)

    if backend not in ("cpython", "rustpython"):
        print(f"Error: unknown backend '{backend}'. Use 'cpython' or 'rustpython'.", file=sys.stderr)
        sys.exit(1)

    print(f"Creating new basilisk project: {project_name} (backend: {backend})")

    # Create directory structure
    src_dir = project_dir / "src"
    src_dir.mkdir(parents=True)

    # Build command depends on backend
    if backend == "rustpython":
        build_cmd = f"BASILISK_PYTHON_BACKEND=rustpython CANISTER_CANDID_PATH=./{project_name}.did python -m basilisk {project_name} src/main.py"
    else:
        build_cmd = f"CANISTER_CANDID_PATH=./{project_name}.did python -m basilisk {project_name} src/main.py"

    dfx_json = f"""\
{{
    "canisters": {{
        "{project_name}": {{
            "type": "custom",
            "build": "{build_cmd}",
            "candid": "{project_name}.did",
            "wasm": ".basilisk/{project_name}/{project_name}.wasm"
        }}
    }}
}}
"""
    (project_dir / "dfx.json").write_text(dfx_json)

    # src/main.py
    main_py = """\
from basilisk import query, update, text, nat64, ic

# A simple counter stored in a global variable.
# State persists across calls but resets on canister upgrade.
counter = 0

@query
def greet(name: text) -> text:
    \"\"\"Return a greeting message.\"\"\"
    return f"Hello, {name}! The counter is at {counter}."

@query
def get_counter() -> nat64:
    \"\"\"Read the current counter value.\"\"\"
    return counter

@update
def increment() -> nat64:
    \"\"\"Increment the counter and return the new value.\"\"\"
    global counter
    counter += 1
    return counter

@query
def get_time() -> nat64:
    \"\"\"Return the current IC timestamp in nanoseconds.\"\"\"
    return ic.time()

@query
def whoami() -> text:
    \"\"\"Return the caller's principal ID.\"\"\"
    return str(ic.caller())
"""
    (src_dir / "main.py").write_text(main_py)

    # .gitignore
    gitignore = """\
.dfx/
.basilisk/
node_modules/
"""
    (project_dir / ".gitignore").write_text(gitignore)

    build_note = "⚡ fast template build" if backend == "cpython" else "🔨 full Rust build (~5-10 min first time)"
    print(f"""
Done! Created {project_name}/ ({build_note})
  src/main.py    — your canister code (query + update examples)
  dfx.json       — IC project config (backend: {backend})

Next steps:
  cd {project_name}
  dfx start --background
  dfx deploy
  dfx canister call {project_name} greet '("World")'
  dfx canister call {project_name} increment
  dfx canister call {project_name} get_counter
""")


def cmd_build():
    """Build the canister in the current directory."""
    # Find dfx.json
    if not Path("dfx.json").exists():
        print("Error: no dfx.json found. Run this from a basilisk project directory.", file=sys.stderr)
        sys.exit(1)

    import json
    with open("dfx.json") as f:
        dfx = json.load(f)

    canisters = dfx.get("canisters", {})
    if not canisters:
        print("Error: no canisters defined in dfx.json.", file=sys.stderr)
        sys.exit(1)

    # Build each canister
    for name, config in canisters.items():
        build_cmd = config.get("build", "")
        if "basilisk" in build_cmd:
            print(f"Building canister: {name}")
            # Extract entry point from build command
            parts = build_cmd.split()
            # Expected: python -m basilisk <name> <entry_point>
            if len(parts) >= 5:
                entry_point = parts[4]
            else:
                entry_point = "src/main.py"

            candid_path = config.get("candid", f"{name}.did")
            os.environ["CANISTER_CANDID_PATH"] = f"./{candid_path}"

            # Run the basilisk build
            import subprocess
            result = subprocess.run(
                [sys.executable, "-m", "basilisk", name, entry_point],
                env={**os.environ},
            )
            if result.returncode != 0:
                print(f"Error: build failed for canister '{name}'.", file=sys.stderr)
                sys.exit(1)


def main():
    if len(sys.argv) < 2:
        print(__doc__.strip())
        sys.exit(1)

    command = sys.argv[1]

    if command == "new":
        # Parse --backend flag
        args = sys.argv[2:]
        backend = "cpython"  # default
        if "--backend" in args:
            idx = args.index("--backend")
            if idx + 1 >= len(args):
                print("Error: --backend requires a value (cpython or rustpython)", file=sys.stderr)
                sys.exit(1)
            backend = args[idx + 1]
            args = args[:idx] + args[idx + 2:]

        if len(args) < 1:
            print("Usage: basilisk new [--backend cpython|rustpython] <project_name>", file=sys.stderr)
            sys.exit(1)
        cmd_new(args[0], backend)

    elif command == "build":
        cmd_build()

    elif command in ("-h", "--help", "help"):
        print(__doc__.strip())

    elif command == "--version":
        from basilisk import __version__
        print(f"basilisk {__version__}")

    else:
        print(f"Unknown command: {command}", file=sys.stderr)
        print(__doc__.strip())
        sys.exit(1)


if __name__ == "__main__":
    main()
